"""The polling mixin for controllers (poller) helps controllers to update the view based on changes in the model"""

from julesTk.controller import Controller

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class Poller(Controller):
    """Poller; a polling controller"""

    def __init__(self, *args, **kwargs):
        super(Poller, self).__init__(*args, **kwargs)
        self._interval = 1  # in seconds
        self._polling = False  # whether the poller is active

    @property
    def interval(self):
        return self._interval

    @interval.setter
    def interval(self, v):
        self._interval = v

    def is_polling(self):
        return self._polling is True

    def set_polling(self, state):
        self._polling = state is True

    def _prepare(self):
        raise NotImplementedError

    def _start(self):
        raise NotImplementedError

    def update(self, observable):
        raise NotImplementedError

    def run(self):
        """Runs the poller"""
        self.set_polling(True)
        self._update()

    def execute(self):
        raise NotImplementedError

    def _update(self):
        if self.is_polling():
            try:
                self.execute()
            except KeyboardInterrupt:
                self.set_polling(False)
            self.view.after(int(self._interval * 1000), self._update)

    def _stop(self):
        self.set_polling(False)
        super(Poller, self)._stop()



"""Provides TopLevel views; i.e. Windows"""

from julesTk.view import tk, BaseView
from viewset import BaseViewSet

__author__ = "Joeri Jongbloets <joeri@jongbloets>"


class Window(tk.Toplevel, BaseView):

    def __init__(self, parent, controller):
        tk.Toplevel.__init__(self, parent)
        BaseView.__init__(self, parent, controller)
        self.protocol("WM_DELETE_WINDOW", self.close)

    def _prepare(self):
        raise NotImplementedError

    def _show(self):
        self.deiconify()

    def _hide(self):
        self.withdraw()
        return True

    def _close(self):
        self.destroy()
        return True


class WindowViewSet(Window, BaseViewSet):
    """A window that can contain multiple views"""

    def _prepare(self):
        raise NotImplementedError

    def _close(self):
        BaseViewSet.close_views(self)
        return super(WindowViewSet, self)._close()



"""Implementing basic view objects"""

import sys
if sys.version_info[0] < 3:
    import Tkinter as tk
    import ttk
else:
    import tkinter as tk
    from tkinter import ttk
from julesTk.app import Application

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class BaseFrame(object):
    """Empty base class used to share most basic functions between frames and views"""

    FONT_FAMILY = "Verdana"
    FONT_SMALL = (FONT_FAMILY, 8)
    FONT_NORMAL = (FONT_FAMILY, 10)
    FONT_LARGE = (FONT_FAMILY, 14)
    FONT_HUGE = (FONT_FAMILY, 16)

    @staticmethod
    def configure_grid(w, sticky="nsew", **kwargs):
        w.grid(sticky=sticky, **kwargs)

    @staticmethod
    def configure_column(w, indexes, weight=1, **kwargs):
        if not isinstance(indexes, (tuple, list)):
            indexes = [indexes]
        kwargs["weight"] = weight
        for index in indexes:
            w.grid_columnconfigure(index, **kwargs)

    @staticmethod
    def configure_row(w, indexes, weight=1, **kwargs):
        if not isinstance(indexes, (tuple, list)):
            indexes = [indexes]
        kwargs["weight"] = weight
        for index in indexes:
            w.grid_rowconfigure(index, **kwargs)


class Frame(ttk.Frame, BaseFrame):

    def __init__(self, parent):
        super(Frame, self).__init__(parent)


class BaseView(BaseFrame):
    """Empty Base class used for checking if an object belongs to the view family"""

    def __init__(self, parent, controller):
        """Initialize a BaseView

        """
        super(BaseView, self).__init__()
        self._configured = False
        self._parent = parent
        self._controller = controller
        self._variables = {}
        self._widgets = {}

    def __del__(self):
        self.close()

    @property
    def parent(self):
        """Return the master of this widget

        :rtype: julesTk.view.BaseView
        """
        return self._parent

    @property
    def application(self):
        """Return the root level widget

        :rtype: Tkinter.Tk or tkinter.Tk
        """
        result = self.parent
        if not isinstance(result, Application):
            result = self.parent.application
        return result

    @property
    def controller(self):
        """ The controller

        :rtype: julesTk.controller.BaseController
        """
        return self._controller

    @property
    def variables(self):
        """Variables registered to this view

        :rtype: dict[str, Tkinter.Variable | tkinter.Variable]
        """
        return self._variables

    @property
    def widgets(self):
        """Widgets registered to this view

        :rtype: dict[str, Tkinter.BaseWidget | tkinter.BaseWidget]
        """
        return self._widgets

    def prepare(self):
        self._prepare()
        self._configured = True
        return self

    def _prepare(self):
        raise NotImplementedError

    def show(self):
        if not self._configured:
            self.prepare()
        return self._show()

    def _show(self):
        raise NotImplementedError

    def hide(self):
        return self._hide()

    def _hide(self):
        raise NotImplementedError

    def close(self):
        return self._close()

    def _close(self):
        raise NotImplementedError

    def has_variable(self, name):
        """Whether a variable is registered under the given name

        :param name: Name to search for
        :type name: str
        :rtype: bool
        """
        return name in self.variables.keys()

    def add_variable(self, name, variable):
        """Register a variable under a new name to the view

        :param name: Name to register the variable under
        :type name: str
        :param variable: The variable to register
        :type variable: Tkinter.Variable | tkinter.Variable
        :rtype: Tkinter.Variable | tkinter.Variable
        """
        if self.has_variable(name):
            raise KeyError("Variable name already registered to this view")
        self.variables[name] = variable
        return variable

    def get_variable(self, name):
        """Get the variable registered to this view under the given name

        :param name: Name of the variable
        :type name: str
        :return: The requested variable
        :rtype: Tkinter.Variable | tkinter.Variable
        """
        if not self.has_variable(name):
            raise KeyError("Variable name not registered to this view")
        return self.variables[name]

    def remove_variable(self, name):
        """Remove a variable from the registry of this view

        The variable is NOT destroyed!

        :param name: Name of the variable to remove
        :type name: str
        :return: The removed variable
        :rtype: Tkinter.Variable | tkinter.Variable
        """
        if not self.has_variable(name):
            raise KeyError("Variable name not registered to this view")
        self.variables.pop(name)
        return not self.has_variable(name)

    def has_widget(self, name):
        """Whether a widget is registered using the given name

        :type name: str
        :rtype: bool
        """
        return name in self.widgets.keys()

    def add_widget(self, name, widget):
        """Register a new widget to this view

        :param name: Name of the widget
        :type name: str
        :param widget: The widget to register
        :type widget: Tkinter.BaseWidget | tkinter.BaseWidget
        :return: The registered widget
        :rtype: Tkinter.BaseWidget | tkinter.BaseWidget
        """
        if self.has_widget(name):
            raise KeyError("Widget name already registered to this view")
        self.widgets[name] = widget
        return widget

    def get_widget(self, name):
        """Get the widget registered under the given name

        :param name: Name of the widget
        :type name: str
        :return: The requested widget
        :rtype: Tkinter.BaseWidget | tkinter.BaseWidget
        """
        if not self.has_widget(name):
            raise KeyError("Widget name not registered to this view")
        return self.widgets[name]

    def remove_widget(self, name):
        """Remove a widget from the registry of this view

        The widget is NOT destroyed

        :param name: Name of the widget to remove
        :type name: str
        :return: The widget that was removed
        :rtype: Tkinter.BaseWidget | tkinter.BaseWidget
        """
        if not self.has_widget(name):
            raise KeyError("Widget name not registered to this view")
        self.widgets.pop(name)
        return not self.has_widget(name)


class View(Frame, BaseView):
    """A Frame acting as """

    def __init__(self, parent, controller):
        Frame.__init__(self, parent)
        BaseView.__init__(self, parent, controller)

    def _show(self):
        """Shows the view"""
        self.tkraise()
        return True

    def _hide(self):
        """Hides the view (temporarily)"""
        self.grid_remove()
        return True

    def _close(self):
        """Closes the view"""
        self.destroy()
        return True

    def _prepare(self):
        """Configure this view"""
        raise NotImplementedError



"""Implement a Frame with a matplotlib"""

from julesTk.view import *

import matplotlib
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, NavigationToolbar2TkAgg
from matplotlib.figure import Figure


class PlotFrame(Frame, object):

    def __init__(self, parent):
        super(PlotFrame, self).__init__(parent)
        self._figure = None
        self._canvas = None
        self._toolbar = None
        self._legend = None
        self._axes = None

    def _setup_figure(self, size, dpi=100):
        if not isinstance(size, tuple) and not len(size) == 2:
            raise ValueError("Invalid value for size (need tuple of length 2)")
        f = Figure(figsize=size, dpi=dpi)
        self._figure = f

    def _setup_canvas(self):
        if not isinstance(self.figure, Figure):
            raise ValueError("Invalid figure object")
        self._canvas = FigureCanvasTkAgg(self.figure, self)
        self._setup_toolbar()
        self.canvas.show()
        self._canvas.get_tk_widget().pack(side=tk.BOTTOM, fill=tk.BOTH, expand=True)

    def _setup_toolbar(self):
        self._toolbar = NavigationToolbar2TkAgg(self.canvas, self)
        self.toolbar.update()
        self.canvas._tkcanvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)

    def _setup_subplot(self):
        self._axes = self.figure.add_subplot(111)

    @property
    def figure(self):
        """Returns the current add_plot figure of this Frame

        :rtype: matplotlib.figure.Figure
        """
        return self._figure

    @property
    def canvas(self):
        """Returns the current canvas of this frame

        :rtype: matplotlib.backends.backend_tkagg.FigureCanvasTkAgg
        """
        return self._canvas

    @property
    def axes(self):
        """Returns the current subplot in the figure

        :rtype: matplotlib.axes.Axes
        """
        return self._axes

    @property
    def legend(self):
        return self._legend

    @property
    def toolbar(self):
        return self._toolbar

    def setup(self, size=None, dpi=100):
        if size is None:
            size = (5, 5)
        self._setup_figure(size, dpi)
        self._setup_canvas()
        self._setup_subplot()

    def add_legend(self):
        if self.axes is not None:
            self._legend = self.axes.legend(loc='best')

    def draw(self):
        self.canvas.draw()

    def clear(self):
        self.figure.clear()
        self._setup_subplot()
        self.canvas.draw()


class PlotView(View):
    """ A view with a plot embedded.

    """

    def __init__(self, parent, controller):
        super(PlotView, self).__init__(parent, controller)
        self._plot = None

    @property
    def plot(self):
        """ Returns the plot frame embedded in this frame

        :rtype: julesTk.view.plot.PlotFrame
        """
        return self._plot

    def body(self):
        self.configure_grid(self)
        self.setup_plot()

    def setup_plot(self):
        self._plot = PlotFrame(self)
        self.plot.setup()



from julesTk.view import View


class BaseViewSet(object):
    """BaseViewSet mixin"""

    def __init__(self):
        super(BaseViewSet, self).__init__()
        self._views = {}

    def __del__(self):
        self.close_views()

    @property
    def views(self):
        """

        :rtype: dict[str, pyLabJackView.view.View]
        """
        return self._views

    def get_view(self, name):
        """

        :param name: Name of the view
        :rtype: julesTk.view.View
        """
        if not self.has_view(name):
            KeyError("No view registered under: {}".format(name))
        return self.views[name]

    def has_view(self, name):
        return name in self.views

    def add_view(self, name, frame):
        if self.has_view(name):
            KeyError("Already registered a view under: {}".format(name))
        self.views[name] = frame

    def remove_view(self, name):
        if not self.has_view(name):
            KeyError("No view registered under: {}".format(name))
        self.get_view(name).close()
        self.views.pop(name)

    def show_view(self, name):
        """Show a specific View"""
        v = self.get_view(name)
        v.show()

    def hide_views(self):
        """Hide all Views"""
        for v in self._views.keys():
            self.hide_view(v)
            # done

    def hide_view(self, name):
        """Hide a specific View"""
        v = self.get_view(name)
        v.hide()

    def close_views(self):
        while len(self.views.values()) > 0:
            name = self.views.keys()[0]
            self.close_view(name)

    def close_view(self, name):
        v = self.get_view(name)
        v.close()
        self.remove_view(name)


class ViewSet(View, BaseViewSet):
    """A view that can contain one or more other views"""

    def __init__(self, parent, controller):
        super(ViewSet, self).__init__(parent, controller)

    def _prepare(self):
        """Configure this ViewSet"""
        raise NotImplementedError

    def _close(self):
        """Close this ViewSet"""
        self.close_views()
        super(ViewSet, self)._close()



"""Classes for creating basic modal dialogs

Such as YES-NO, Question, Information dialogs etc.

"""

from julesTk.controller.window import WindowController
from julesTk import view, controller
from julesTk.view.window import Window

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class ModalWindow(Window):

    def __init__(self, parent, controller):
        super(ModalWindow, self).__init__(parent, controller)
        self.application.register_hook("APP_CLOSE", self.hide)

    def _prepare(self):
        raise NotImplementedError

    def _show(self):
        super(ModalWindow, self)._show()
        self.transient(self.parent)
        self.grab_set()
        self.application.wait_window(self)

    def _hide(self):
        return False

    def _close(self):
        self.application.remove_hook("APP_CLOSE", self.hide)
        super(ModalWindow, self)._close()


class Dialog(ModalWindow):

    def __init__(self, parent, controller):
        super(Dialog, self).__init__(parent, controller)
        self._response = None

    @property
    def response(self):
        """Returns the input of the user given in the ModalWindow

        Developers can use this communicate the input of the window to the controller
        """
        return self._response

    def _prepare(self):
        self.grid()
        self.configure_column(self, 0)
        self.configure_row(self, [0, 1, 2])
        # header
        fmh = self.add_widget(
            "header", view.ttk.Frame(self)
        )
        self.header(fmh)
        self.configure_grid(fmh, row=0, column=0)
        # body
        fmb = self.add_widget(
            "body", view.ttk.Frame(self)
        )
        self.body(fmb)
        self.configure_grid(fmb, row=1, column=0)
        # footer
        fmf = self.add_widget(
            "footer", view.ttk.Frame(self)
        )
        self.footer(fmf)
        self.configure_grid(fmf, row=2, column=0)

    def header(self, parent):
        """Header of the dialog"""
        return True  ## override

    def body(self, parent):
        """Build the body of the dialog, parent refers to parent frame"""
        return True  ## override

    def footer(self, parent):
        """Build the buttons of the dialog, parent refers to parent frame"""
        return True  ## override

    def validate(self):
        return True  ## override

    def start(self):
        self.show()


class MessageBox(Dialog):

    def __init__(self, parent, controller, buttons=None):
        """ Initialize a MessageBox

        :param parent:
        :type parent:
        :param controller:
        :type controller:
        :param buttons: List of button definitions.
            A button definition is dictionary with the keys: id, caption, value
        :type buttons: list[dict[str, str | int | float]]
        """
        super(MessageBox, self).__init__(parent, controller)
        self._message = view.tk.StringVar("")
        if buttons is None:
            buttons = []
        if len(buttons) == 0:
            buttons = [{"id": "ok", "caption": "Ok", "value": True}]
        self._buttons = buttons
        self._functions = {}

    @classmethod
    def alert(cls, parent, title, message, buttons=None):
        """Show an alert"""
        if not isinstance(parent, controller.BaseController):
            raise ValueError("Expected a controller not a {}".format(type(parent)))
        mb = cls(parent.view, parent, buttons=buttons)
        mb.title = title
        mb.message = message
        mb.show()

    @property
    def message(self):
        return self._message.get()

    @message.setter
    def message(self, value):
        self._message.set(value)

    def body(self, parent):
        lbm = view.ttk.Label(parent, textvariable=self._message)
        self.configure_grid(lbm, padx=10, pady=5)

    def footer(self, parent):
        idx = 0
        for col, button in enumerate(self._buttons):
            btn = None
            # get button id
            id_ = button.get("id", None)
            if id_ is None:
                id_ = idx
                idx += 1
            # get caption
            caption = button.get("caption", id_)
            # get return value
            value = button.get("value", id_)
            # check if set to default
            default = button.get("default", False)
            if default:
                self._response = value
            # add button
            btn = view.ttk.Button(
                parent, text=caption,
                default=view.tk.ACTIVE if default else view.tk.NORMAL,
                command=lambda i=value: self.process_click(i)
            )
            self.add_widget(id_, btn)
            btn.pack(side=view.tk.LEFT, padx=5)

    def process_click(self, value):
        self._response = value
        self.close()


class QuestionBox(Dialog):

    def __init__(self, parent, controller):
        super(QuestionBox, self).__init__(parent, controller)
        self._question = view.tk.StringVar(self)
        self._answer = view.tk.StringVar(self)
        self._error = view.tk.StringVar(self)

    @classmethod
    def ask(cls, parent, question, default=None):
        if not isinstance(parent, controller.BaseController):
            raise ValueError("Expected a controller not a {}".format(type(parent)))
        qb = cls(parent.view, parent)
        qb.question = question
        qb._response = default
        qb.answer = default
        qb.show()
        return qb.response

    @property
    def question(self):
        return self._question.get()

    @question.setter
    def question(self, value):
        self._question.set(value)

    @property
    def answer(self):
        return self._answer.get()

    @answer.setter
    def answer(self, value):
        value = "" if value is None else value
        self._answer.set(value)

    @property
    def error(self):
        return self._error

    @error.setter
    def error(self, text):
        self._error.set(text)

    def show_validation_msg(self):
        lbv = self.get_widget("validate")
        lbv.grid()

    def hide_validation_msg(self):
        lbv = self.get_widget("validate")
        lbv.grid_remove()

    def header(self, parent):
        # add question
        lbq = view.ttk.Label(parent, textvariable=self._question)
        self.configure_grid(
            lbq, padx=10, pady=5
        )

    def body(self, parent):
        # add answer
        ena = view.ttk.Entry(
            parent, textvariable=self._answer
        )
        self.configure_grid(
            ena, padx=15, pady=5
        )
        # add validation
        view.ttk.Style().configure(
            "Error.TLabel", foreground="red"
        )
        lbv = view.ttk.Label(
            parent, textvariable=self._error, style="Error.TLabel"
        )
        self.add_widget("validate", lbv)
        self.configure_grid(
            lbv, row=1, padx=20
        )
        self.hide_validation_msg()

    def footer(self, parent):
        self.configure_column(parent, [0, 1])
        # add cancel
        view.ttk.Button(
            parent, text="Cancel", command=self.cancel
        ).pack(side=view.tk.LEFT)
        self.bind("<Escape>", lambda x: self.cancel())
        # add ok
        view.ttk.Button(
            parent, text="Ok", command=self.ok
        ).pack(side=view.tk.LEFT)
        self.bind("<Return>", lambda x: self.ok())

    def validate(self):
        response = self._answer.get()
        result = response not in (None, "")
        if not result:
            self.error = "Please provide an answer"
            self.show_validation_msg()
        return result

    def cancel(self):
        self.close()

    def ok(self):
        if self.validate():
            self._response = self._answer.get()
            self.close()



"""Provide Controller classes"""

from julesTk.app import Application
from julesTk.model import Model
from julesTk.view import BaseView, View, tk

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class BaseController(object):
    """A bare-bones controller, without view or model"""

    def __init__(self, parent):
        super(BaseController, self).__init__()
        self._parent = parent
        self._configured = False

    def __del__(self):
        self.stop()

    def prepare(self):
        self._prepare()
        self._configured = True
        return self

    def _prepare(self):
        """Configures the controller (set-up model, view)"""
        raise NotImplementedError

    def start(self):
        """Start the controller"""
        if not self._configured:
            self.prepare()
        return self._start()

    def _start(self):
        raise NotImplementedError

    def stop(self):
        """Stop the controller"""
        return self._stop()

    def _stop(self):
        raise NotImplementedError

    @property
    def parent(self):
        """Return reference to the parent object

        :rtype: julesTk.controller.BaseController or julesTk.app.Application
        """
        return self._parent

    @property
    def application(self):
        """Return the application instance

        :rtype: julesTk.app.Application
        """
        result = self.parent
        if not isinstance(result, tk.Tk):
            result = self.parent.application
        return result


class ViewController(BaseController):
    """A controller with a view (no model)"""

    VIEW_CLASS = View

    def __init__(self, parent, view=None):
        super(ViewController, self).__init__(parent=parent)
        self._view = view

    @property
    def view(self):
        """Return the view managed by this controller

        :rtype: julesTk.view.View
        """
        return self._view

    @property
    def parent_view(self):
        """Return the view of the parent of this controller

        :rtype: julesTk.view.View
        """
        pview = self.parent
        if isinstance(pview, BaseController):
            pview = self.parent.view
        return pview

    def has_view(self):
        """Whether the controller has a view attached"""
        return self.view is not None

    def _start(self):
        """Start the controller and open the view"""
        self.view.show()

    def _stop(self):
        """Stop the controller and close the view managed by the view"""
        self.view.close()

    def _prepare(self):
        """Configure the controller and view.

        Should be called before showing the view or starting the controller.
        Preferably after initialization of the controller.

        :rtype: julesTk.controller.ViewController
        """
        if not self.has_view():
            self.load_view()
        self.view.prepare()
        return self

    def load_view(self):
        """Returns a new instance of the view as defined in VIEW_CLASS

        :return:
        :rtype: julesTk.view.View
        """
        if not issubclass(self.VIEW_CLASS, BaseView):
            raise ValueError("Expected a view not {}".format(self.VIEW_CLASS.__name__))
        self._view = self.VIEW_CLASS(self.parent_view, self)
        return self.view


class ModelController(BaseController):
    """A controller with a model (no view)"""

    def __init__(self, parent, model=None):
        super(ModelController, self).__init__(parent=parent)
        self._model = model

    @property
    def model(self):
        """The model used by this controller

        :return:
        :rtype: julesTk.model.Model
        """
        return self._model

    @model.setter
    def model(self, model):
        self._set_model(model)

    def _set_model(self, model):
        if not isinstance(model, Model):
            raise ValueError("Expected a model class")
        self._model = model

    def has_model(self):
        """Whether the controller has a model attached"""
        return self.model is not None

    def _start(self):
        """Start the controller"""
        raise NotImplementedError

    def _stop(self):
        """Stop the controller"""
        raise NotImplementedError

    def _prepare(self):
        """Set-up and configure the controller"""
        raise NotImplementedError


class Controller(ViewController, ModelController):
    """A controller with both a model and view"""

    def __init__(self, parent, view=None, model=None):
        ViewController.__init__(self, parent=parent, view=view)
        ModelController.__init__(self, parent=parent, model=model)




from . import ViewController
from julesTk.view.window import Window

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class WindowController(ViewController):

    VIEW_CLASS = Window

    def load_view(self):
        if not issubclass(self.VIEW_CLASS, Window):
            raise ValueError("Expected a ModalWindow not {}".format(self.VIEW_CLASS.__name__))
        self._view = self.VIEW_CLASS(self.parent_view, self)



"""Module managing starting and stopping the application

"""

import sys
if sys.version_info[0] < 3:
    import Tkinter as tk
else:
    import tkinter as tk
from . import ThreadSafeObject

__author__ = "Joeri Jongbloets <joeri@jongbloets.net>"


class Application(ThreadSafeObject, tk.Tk):
    """Application entry point"""

    def __init__(self, *args, **kwargs):
        tk.Tk.__init__(self, *args, **kwargs)
        super(Application, self).__init__()
        self._configured = False
        self.protocol("WM_DELETE_WINDOW", self.stop)
        self._controllers = {}
        self._hooks = {}

    @property
    def controllers(self):
        """All controllers attached to this application

        :return: A dictionary of the name - controller mapping
        :rtype: dict[str, julesTk.controller.BaseController]
        """
        return self._controllers

    def get_controller(self, name):
        """Returns the controller registered under the given name

        :raises KeyError: If no controller is registered with that name

        :param name: Name of the controller
        :type name: str
        :rtype: julesTk.controller.BaseController
        """
        if not self.has_controller(name):
            raise KeyError("No controller registered under: {}".format(name))
        return self.controllers[name]

    def has_controller(self, name):
        """Returns whether a controller is registered under the given name

        :param name: Name of the controller
        :type name: str
        :rtype: bool
        """
        return name in self.controllers.keys()

    def add_controller(self, name, controller):
        """Registers a new controller under the given name

        :raises KeyError: If another controller is already registered under the same name

        :param name: Name to register the controller under
        :type name: str
        :param controller: The controller to register
        :type controller: julesTk.controller.BaseController
        """
        if self.has_controller(name):
            raise KeyError("Already registered a controller under: {}".format(name))
        self.controllers[name] = controller

    def remove_controller(self, name):
        """Removes a controller from the application

        :raises KeyError: If no controller is registered with that name
        :param name: Name of the controller
        :type name: str
        """
        if not self.has_controller(name):
            raise KeyError("No controller registered under: {}".format(name))
        self.controllers.pop(name)

    def has_hook(self, name):
        return name in self._hooks.keys()

    def register_hook(self, name, f):
        if self.has_hook(name):
            self._hooks[name].append(f)
        else:
            self._hooks[name] = [f]
        return self

    def remove_hook(self, name, f):
        if self.has_hook(name):
            hooks = self._hooks[name]
            if f in hooks:
                hooks.remove(f)
        return self

    def process_hook(self, name):
        hooks = []
        if self.has_hook(name):
            hooks = self._hooks[name]
        result = True
        for f in hooks:
            result = f() and result
        return result

    def prepare(self):
        self._prepare()
        self._configured = True
        return self

    def _prepare(self):
        """Configures the application and loads at least one controller"""
        raise NotImplementedError

    def run(self):
        """Start the application"""
        with self.lock:
            # _prepare the application
            self.prepare()
            # now start the application
            self.start()
        # in the main loop we wait, so we do not need a lock
        self.mainloop()
        with self.lock:
            # clean up
            self.stop()

    def start(self):
        if not self._configured:
            self.prepare()
        self._start()

    def _start(self):
        """Everything to show and run the application"""
        raise NotImplementedError

    def stop(self):
        # process app close hook
        if self.process_hook("APP_CLOSE"):
            self._stop()

    def _stop(self):
        """Clean-up after execution"""
        while len(self.controllers) > 0:
            name = self.controllers.keys()[0]
            self.get_controller(name).stop()
            self.remove_controller(name)
        tk.Tk.quit(self)
